import json

from .feature import parse


class BedTable:

    # Always remember the *self* argument
    def __init__(self, bed_file):

        self.features = []

        featureList = parse(bed_file)
        unique_id = 1
        for var in featureList:
            self.features.append((var, unique_id))
            unique_id += 1

    def to_JSON(self):

        jsonArray = [];

        for tuple in self.features:
            feature = tuple[0]
            unique_id = tuple[1]
            obj = {
                "unique_id": unique_id,
                "Chrom": feature.chr,
                "Start": feature.start + 1,
                "End": feature.end,
                "Name": feature.name
            }

            jsonArray.append(obj)

        return json.dumps(jsonArray)



import gzip
import io
import pysam
import requests
from intervaltree import IntervalTree


class Feature:

    def __init__(self, chr, start, end, text, name=''):
        self.chr = chr
        self.start = start
        self.end = end
        self.text = text
        self.name = name


class FeatureReader:

    def __init__(self, path):

        tabix = None
        if path.endswith(".gz"):
            # Might be a tabix file
            try:
                tabix = pysam.TabixFile(path)
            except:
                tabix = None

        if tabix:
            self.reader = _Tabix(tabix)
        else:
            self.reader = _NonIndexed(path)

    def slice(self, region=None):
        return self.reader.slice(region)


class _Tabix:

    def __init__(self, tabix):
        self.file = tabix

    def slice(self, region=None):

        tb = self.file
        if region:
            range_string = region['chr'] + ":" + str(region['start']) + "-" + str(region['end'])
            it = tb.fetch(range_string)
        else:
            it = tb.fetch()

        data = ""
        for row in it:
            data += row + '\n'

        return data


## Implement a pysam/tabix style inteface for non-indexed files
class _NonIndexed:

    def __init__(self, file):

        self.file = file
        self.tree = None

    def slice(self, region=None):

        if region == None:
            f = None
            try:
                f = getstream(self.file)
                return f.read()
            finally:
                if f:
                    f.close()

        else:
            if not self.tree:
                features = parse(self.file)
                self.tree = FeatureTree(features)

            reference = region["chr"]
            start = region["start"]
            end = region["end"]
            feature_intervals = self.tree.query(reference, start, end)

            features = []
            if feature_intervals:
                features = []
                for i in feature_intervals:
                    features.append(i.data)

                def sortFunc(f):
                    return f.start
                features = sorted(features, key=sortFunc)


        content = ''
        for data in features:
            content += data.text

        return content



class FeatureTree:

    def __init__(self, featureList):

        self.featureMap = dict();

        for f in featureList:
            chr = f.chr
            tree = self.featureMap.get(chr)
            if tree == None:
                tree = IntervalTree()
                self.featureMap[chr] = tree
            tree[f.start:f.end] = f

    def query(self, chr, start, end):
        tree = self.featureMap.get(chr)
        if tree == None:
            return set()
        else:
            return tree[start:end]


def get_data(filename, region=None):
    reader = FeatureReader(filename)
    return reader.slice(region)

def parse(path, format=None):
    '''
    Parse a feature file and return an array of feature objects.  Supported formats are bed, gff, and gtf.
    :param path: Path to feature file, which can be local or url
    :param format: File format, bed | gtf | gff
    :return: List of feature objects {chr, start, end, text, name}
    '''

    f = None
    try:
        f = getstream(path)
        if not format:
            format = infer_format(path)
        if format == 'bed':
            return parse_bed(f)
        elif format == 'gff' or format == 'gtf':
            return parse_gff(f)
        else:
            raise Exception("Unknown file format: " + path)
    finally:
        if f:
            f.close()

def parse_bed(f):
    features = []
    for line in f:
        if not (line.startswith('#') or line.startswith('track') or line.startswith('browser')):
            tokens = line.rstrip('\n').rstrip('\r').split('\t')
            chr = tokens[0]
            start = int(tokens[1])
            end = int(tokens[2])
            name = tokens[3] if len(tokens) > 3 else ''
            features.append(Feature(chr, start, end, line, name))
    return features

def parse_gff(f):
    features = []
    for line in f:
        if not (line.startswith('#') or line.startswith('track') or line.startswith('browser')):
            tokens = line.rstrip('\n').rstrip('\r').split('\t')
            chr = tokens[0]
            start = int(tokens[3]) - 1
            end = int(tokens[4])
            name = ''
            features.append(Feature(chr, start, end, line, name))

    return features

def infer_format(filename):
    '''
    Infer the genomic file format from the filename.  First known formats are checked.  Next presenece of
    the magic string "refgene" in the filename is checked for UCSC refgene files.  This is a legacy
    IGV convention.  The order is important, a recognized extension wins.
    NOTE: Formats are for output data uris.  CRAM format is converted to BAM before output.
    :param filename:
    :return:
    '''
    filename = filename.lower()
    if (filename.endswith(".gz")):
        filename = filename[:-3]

    if filename.endswith(".bam"):
        return "bam"
    if filename.endswith(".cram"):
        return "bam"
    elif filename.endswith(".vcf"):
        return "vcf"
    elif filename.endswith(".bed"):
        return "bed"
    elif filename.endswith(".gff") or filename.endswith(".gff3"):
        return "gff"
    elif filename.endswith(".gtf"):
        return "gtf"
    elif filename.find("refgene"):
        return "refgene"
    else:
        idx = filename.rfind(".")
        if idx > 0:
            return filename[idx + 1:]
        else:
            return None
        idx = filename.rfind(".")
        if idx > 0:
            return filename[idx + 1:]
        else:
            return None

def getstream(file):

    # TODO -- gcs

    if file.startswith('http://') or file.startswith('https://'):
        response = requests.get(file)
        status_code = response.status_code    #TODO Do something with this

        if file.endswith('.gz'):
            content = response.content
            text = gzip.decompress(content).decode('utf-8')
        else:
            text = response.text
        f = io.StringIO(text)
        return text

    elif file.endswith('.gz'):
        f = gzip.open(file, mode='rt')

    else:
        f = open(file, encoding='UTF-8')

    return f



import os
import sys
import json
import math
import argparse
from urllib.request import urlopen
from igv_reports import fasta, ideogram, datauri, tracks, feature, bam, vcf, utils
from igv_reports.varianttable import VariantTable
from igv_reports.bedtable import BedTable


def create_report(args):

    variants_file = args.sites

    if variants_file.endswith(".vcf") or variants_file.endswith (".vcf.gz"):
        table = VariantTable(variants_file, args.species, args.info_columns, args.sample_columns)

    elif variants_file.endswith(".bed") or variants_file.endswith(".bed.gz"):
        table = BedTable(variants_file)

    table_json = table.to_JSON()

    session_dict = {}

    # Create file readers for tracks.  This is done outside the loop so initialization happens onc
    trackreaders = []
    if args.tracks is not None:
        for track in args.tracks:
            reader = utils.getreader(track)
            trackreaders.append({
                "track": track,
                "reader": reader
            })


    # loop through variants creating an igv.js session for each one
    for tuple in table.features:

        feature = tuple[0]
        unique_id = tuple[1]

        # Define a genomic region around the variant
        chr = feature.chr
        position = int(math.floor((feature.start + feature.end) / 2)) + 1   # center of region in 1-based coordinates
        start = int (math.floor(feature.start - int(args.flanking) / 2))
        end = int (math.ceil(feature.end + int(args.flanking) / 2))
        region = {
            "chr": chr,
            "start": start,
            "end": end
        }

        # Fasta
        data = fasta.get_data(args.fasta, region)
        fa = '>' + chr + ':' + str(start) + '-' + str(end) + '\n' + data
        fasta_uri = datauri.get_data_uri(fa)
        fastaJson = {
            "fastaURL": fasta_uri,
        }

        # Ideogram
        if(args.ideogram):
            ideo_string = ideogram.get_data(args.ideogram, region)
            ideo_uri = datauri.get_data_uri(ideo_string)
            fastaJson["cytobandURL"] = ideo_uri


        # Initial locus, +/- 20 bases
        initial_locus = chr + ":" + str(position - 20) + "-" + str(position + 20)
        session_json = {
            "locus": initial_locus,
            "reference": fastaJson,
            "tracks": []
        }

        for tr in trackreaders:

            track = tr["track"]
            reader = tr["reader"]
            trackObj = tracks.get_track_json_dict(track)
            data = reader.slice(region)
            trackObj["url"] = datauri.get_data_uri(data)
            if(trackObj["type"] == "alignment"):
                trackObj["height"] = 500

                # Sort TODO -- do this only for SNV
                # if (trackObj["type"]) == "alignment":
                #     trackObj["sort"] = {
                #         "option": "NUCLEOTIDE",
                #         "locus": chr + ":" + str(variant.pos - 1)
                #     }

            session_json["tracks"].append(trackObj)


        # Build the session data URI

        session_string = json.dumps(session_json);

        session_uri = datauri.get_data_uri(session_string)

        session_dict[str(unique_id)] = session_uri

    session_dict = json.dumps(session_dict)

    template_file = args.template
    if None == template_file:
        template_file = os.path.dirname(sys.modules['igv_reports'].__file__) + '/templates/variant_template.html'

    output_file = args.output

    standalone = args.standalone
    with open(template_file, "r") as f:
        data = f.readlines()

        with open(output_file, "w") as o:

            for i, line in enumerate(data):

                if standalone and "<script" in line and line.find(".js") > 0:
                    inline_script(line, o)

                else:
                    j = line.find('"@TABLE_JSON@"')
                    if j >= 0:
                        line = line.replace('"@TABLE_JSON@"', table_json)

                    j = line.find('"@SESSION_DICTIONARY@"')
                    if j >= 0:
                        line = line.replace('"@SESSION_DICTIONARY@"', session_dict)

                    o.write(line)


def inline_script(line, o):
    #<script type="text/javascript" src="https://igv.org/web/test/dist/igv.min.js"></script>
    s = line.find('src="')
    if s > 0:
        e = line.find('">', s)
        url = line[s+5:e]

        response = urlopen(url)
        js = response.read().decode('utf-8')
        response.close()

        o.write('<script type="text/javascript">\n')
        o.write(js)
        o.write('</script>\n')



def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("sites", help="vcf file defining variants, required")
    parser.add_argument("fasta", help="reference fasta file, required")
    parser.add_argument("--ideogram", help="ideogram file in UCSC cytoIdeo format")
    parser.add_argument("--tracks", nargs="+", help="list of track files")
    parser.add_argument("--template", help="html template file", default=None)
    parser.add_argument("--output", help="output file name", default="igvjs_viewer.html")
    parser.add_argument("--info-columns", nargs="+", help="list of VCF info field names to include in variant table")
    parser.add_argument("--sample-columns", nargs="+", help="list of VCF sample/format field names to include in variant table")
    parser.add_argument("--species", help="Latin species name with underscore as used by ensembl.org", default='homo_sapiens')
    parser.add_argument("--flanking", help="genomic region to include either side of variant", default=1000)
    parser.add_argument('--standalone', help='Print more data', action='store_true')
    args = parser.parse_args()
    create_report(args)

if __name__ == "__main__":

    main()



import unittest
import pathlib
from igv_reports import varianttable, bedtable

class TableTest(unittest.TestCase):

    def test_get_header(self):

        vcf_path = str((pathlib.Path(__file__).parent / "data/minigenome/variants.vcf").resolve())
        table = varianttable.VariantTable(vcf_path)
        self.assertEqual(len(table.variants), 28)

    def test_bedtable(self):

        bed_file = str((pathlib.Path(__file__).parent / "data/minigenome/variants.bed").resolve())
        table = bedtable.BedTable(bed_file)

        json = table.to_JSON()
        self.assertEqual(len(table.features), 4)
        self.assertTrue(json)

    def test_gfftable(self):

        gff = str((pathlib.Path(__file__).parent / "data/minigenome/annotations.gtf.gz").resolve())
        table = bedtable.BedTable(gff)
        json = table.to_JSON()
        self.assertTrue(len(table.features), 2106)
        self.assertTrue(json)









import io
import json
import pysam
from .feature import Feature

class VariantTable:

    # Always remember the *self* argument
    def __init__(self, vcfFile, species, info_columns = None, sample_columns = None):

        vcf = pysam.VariantFile(vcfFile)

        self.info_fields =  info_columns or []
        self.sample_fields = sample_columns or []
        self.variants = []
        self.features = []   #Bed-like features
        self.species = species

        for unique_id, var in enumerate(vcf.fetch()):
            self.variants.append((var, unique_id))
            chr = var.chrom
            start = var.pos - 1
            end = start + 1       #TODO -- handle structure variants and deletions > 1 base
            self.features.append((Feature(chr, start, end, ''), unique_id))

    def to_JSON(self):


        json_array = [];

        for variant, unique_id in self.variants:
            obj = {
                'unique_id': unique_id,
                'CHROM': variant.chrom,
                'POSITION': variant.pos,
                'REF': variant.ref,
                'ALT': ','.join(variant.alts),
                'ID': ''
            }

            if variant.id is not None:
                obj['ID'] = render_ids(variant.id)

            for h in self.info_fields:
                v = ''
                if h in variant.info:
                    if h == 'ANN':
                        genes, effects, impacts, transcript, aa_alt, nt_alt = decode_ann(variant, self.species)
                    elif h == 'COSMIC_ID':
                        v = render_id(v)
                    else:
                        v = render_values(variant.info[h])
                if h == 'ANN':
                    obj['GENE'] = genes
                    obj['EFFECTS'] = effects
                    obj['IMPACT'] = impacts
                    obj['TRANSCRIPT'] = transcript
                    obj['PROTEIN ALTERATION'] = aa_alt
                    obj['DNA ALTERATION'] = nt_alt
                else:
                    obj[h] = v

            for h in self.sample_fields:
                for sample, values in variant.samples.items():
                    v = ''
                    try:
                        v = values[h]
                    except KeyError:
                        # ignore if key is not present
                        pass

                    obj[f'{sample}:{h}'] = render_values(v)

            json_array.append(obj)

        if not any(obj['ID'] for obj in json_array):
            # Remove ID column if none of the records actually had an ID.
            for obj in json_array:
                del obj['ID']
        return json.dumps(json_array)


def render_value(v):
    """Render given value to string."""
    if isinstance(v, float):
        # ensure that we don't waste space by insignificant digits
        return f'{v:.2g}'
    return str(v)


def render_values(v):
    if isinstance(v, str) or isinstance(v, int) or isinstance(v, float):
        return render_value(v)
    return ','.join(map(render_value, v))


def render_id(v):
    if v.startswith('COSM'):
        return (f'<a href = "https://cancer.sanger.ac.uk/cosmic/mutation/overview?'
                f'id={v[4:]}" target="_blank">{v}</a>')
    return v


def render_ids(v):
    return ','.join(map(render_id, v.split(';')))


def decode_ann(variant, species):
    """Decode the standardized ANN field to something human readable."""
    annotations = ([variant.info['ANN'].split('|'
                   )] if isinstance(variant.info['ANN'],
                   str) else [e.split('|') for e in variant.info['ANN']])
    genes = []
    effects = []
    impacts = []
    transcripts = []
    aa_alts = []
    nt_alts = []
    for allele in variant.alts:
        for ann in annotations:
            ann_allele, kind, impact, gene = ann[:4]
            feature_id = ann[6]
            nt_mod, aa_mod = ann[9:11]

            if allele != ann_allele:
                continue

            full = '|'.join(ann)
            # Keep the most severe effect.
            # Link out to Genecards and show the full record in a tooltip.
            genes.append(f'<a href="https://www.ensembl.org/{species}/Gene/'
                           f'Summary?db=core;t={gene}" target="_blank">{gene}</a>')
            effects.append(kind.replace('&', '/'))
            impacts.append(impact)
            transcripts.append(f'<a href="https://www.ensembl.org/{species}/Transcript/Summary?db=core;t={feature_id}" target="_blank">{feature_id}</a>')
            aa_alts.append(f'<a href="https://www.google.com/search?q={gene}&as_epq={aa_mod}" target="_blank">{aa_mod}</a>')
            nt_alts.append(f'<a href="https://www.google.com/search?q={gene}&as_epq={nt_mod}" target="_blank">{nt_mod}</a>')
    return ','.join(genes), ','.join(effects), ','.join(impacts), ','.join(transcripts), ','.join(aa_alts), ','.join(nt_alts)




from base64 import b64encode
from gzip import compress
import argparse
from . import utils
from .regions import parse_region


#This module exports functions to convert text or binary data to a data URI readable by igv.js.

def get_data_uri(data):

    """
    Return a data uri for the input, which can be either a string or byte array
    """

    if isinstance(data, str):
        data = compress(data.encode())
        mediatype = "data:application/gzip"
    else:
        if data[0] == 0x1f and data[1] == 0x8b:
            mediatype = "data:application/gzip"
        else:
            mediatype = "data:application:octet-stream"

    enc_str = b64encode(data)

    data_uri = mediatype + ";base64," + str(enc_str)[2:-1]
    return data_uri


def file_to_data_uri(filename, filetype=None, genomic_range=None):
    reader = utils.getreader(filename, filetype)
    region = parse_region(genomic_range) if genomic_range else None
    data = reader.slice(region)
    data_uri = get_data_uri(data)
    return data_uri


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="name of file to be converted to data uri")
    parser.add_argument("-t", "--filetype", help="type of file to be converted to data uri")
    parser.add_argument("-r", "--region" , help="genomic region to be converted in the form chr:start-stop")
    args = parser.parse_args()

    if args.region:
        region = parse_region(args.region)
    else:
        region = None

    uri = file_to_data_uri(args.filename, args.filetype, region)
    print(uri)

if __name__ == "__main__":
    main()



import unittest
import pathlib
import io
from igv_reports.feature import FeatureReader, _NonIndexed, parse_gff
from igv_reports import datauri

class FeatureFileTest(unittest.TestCase):


    def test_query(self):

        gff = str((pathlib.Path(__file__).parent / "data/minigenome/annotations.gtf.gz").resolve())
        range =  "minigenome:5000-6000"
        uri = datauri.file_to_data_uri(gff, 'gff', range)
        self.assertIsNotNone(uri)



    def test_noquery(self):

        file = str((pathlib.Path(__file__).parent / "data/minigenome/variants.bed").resolve())
        uri = datauri.file_to_data_uri(file)
        self.assertIsNotNone(uri)





import unittest

from igv_reports import regions


class RegionsTest(unittest.TestCase):

    def test_merge_regions(self):

        rlist = ["chr1:1000-2000", "chr1:1005-3000", "chr1:2000-2500", "chr1:50000-60000", "chr2:200"]

        merged_regions = regions.merge_regions(rlist)

        self.assertEqual(3, len(merged_regions))

        m = merged_regions[0]
        self.assertEqual('chr1', m['chr'])
        self.assertEqual(1000, m['start'])
        self.assertEqual(3000, m['end'])
